/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.bus.logicalClock;

import org.objectweb.dream.message.Message;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;

/**
 * Component interface to manipulate logical clock. A logical clock protocol
 * allows <i>n</i> processes to exchange ordered messages. Each process has a
 * identifier form <i>0 </i> to <i>n-1 </i>. When a process send a message to an
 * other, this message is stamped with a logical clock. Before to be delivered
 * to the destination process, the stamp of a given message is analyzed to know
 * if other messages should be delivered before this one.
 */
@Interface(name = "logical-clock")
public interface LogicalClock
{
  /** The default name of {@link LogicalClock }interfaces */
  String LOGICAL_CLOCK_ITF_NAME = "logical-clock";

  /**
   * the value returned by {@link #testRecvMessage(Message)} if the message can
   * be delivered
   */
  int    DELIVER                = 0;

  /**
   * the value returned by {@link #testRecvMessage(Message) }if there is other
   * messages that must be delivered before this one.
   */
  int    WAIT_TO_DELIVER        = 1;

  /**
   * the value returned by {@link #testRecvMessage(Message) }if the message has
   * already been delivered.
   */
  int    ALREADY_DELIVERED      = 2;

  /**
   * Test if the given message can be delivered. If the message is ready to be
   * delivered, the method returns <code>DELIVER</code>. If the message has
   * already been delivered, the method returns <code>ALREADY_DELIVERED</code>,
   * and if other messages are waited before this message the method returns
   * <code>WAIT_TO_DELIVER</code>.
   * 
   * @param message the message to test.
   * @return <code>DELIVER</code>,<code>ALREADY_DELIVERED</code>, or
   *         <code>WAIT_TO_DELIVER</code> code.
   */
  int testRecvMessage(Message message);

  /**
   * Stamps the given message.
   * 
   * @param message a message to stamp
   * @param to the address of the destination process.
   */
  void stamp(Message message, int to);

}
