/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.channel;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExceptionChunk;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implementation of an export multiplexer protocol. This protocol allows to
 * export multiple {@link ChannelFactory} using a single lower level
 * <code>Channel</code>.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
@Provides(interfaces = {@Interface(name = MultiplexExportProtocol.ITF_NAME, signature = MultiplexExportProtocol.class)})
public class MultiplexExportProtocolImpl
    implements
      MultiplexExportProtocol,
      ChannelFactory
{

  private static final String                   BIND_CHUNK_NAME       = "multiexp-bind";
  private static final String                   BIND_REPLY_CHUNK_NAME = "multiexp-bind-reply";
  private ExportIdentifier                      lowerExportId         = null;
  private Map<Short, ChannelFactory>            exportedChanels       = new HashMap<Short, ChannelFactory>();
  private ChunkFactoryReference<MultiplexChunk> multiplexChunkFactory;
  private ChunkFactoryReference<ExceptionChunk> exceptionChunkFactory;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  @Requires(name = ChannelProtocol.LOWER_PROTOCOL_ITF_NAME)
  private ChannelProtocol                       lowerLevelProtocolItf;
  @Requires(name = "message-manager")
  private MessageManagerType                    messageManagerItf;

  // --------------------------------------------------------------------------
  // Services interfaces
  // --------------------------------------------------------------------------

  /**
   * Component reference
   */
  @Service
  Component                                     weaveableC;

  /**
   * Logger of the component
   */
  @DreamMonolog()
  protected Logger                              logger;

  // ---------------------------------------------------------------------------
  // Implementation of the Channel interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChannelFactory#instantiate(OutgoingPush)
   */
  public IncomingPush instantiate(OutgoingPush toClientPush)
      throws BindException
  {
    return new ServerSessionImpl(toClientPush);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Protocol interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChannelProtocol#bind(ExportIdentifier, IncomingPush, Map)
   */
  public OutgoingPush bind(ExportIdentifier exportId,
      IncomingPush toClientPush, Map<String, Object> hints)
      throws InvalidExportIdentifierException, BindException
  {
    // TODO test local binding ie. nextExportID == lowerExportId
    if (!(exportId instanceof MultiplexExportIdentifier))
    {
      throw new InvalidExportIdentifierException(
          "this protocol handle only MultiplexExportIdentifier", exportId);
    }
    MultiplexExportIdentifier meid = (MultiplexExportIdentifier) exportId;
    ClientSessionImpl session = new ClientSessionImpl(toClientPush);

    // propagate binding
    OutgoingPush outgoingPush = lowerLevelProtocolItf.bind(meid
        .getNextExportIds()[0], session, hints);
    session.lowerOutgoingPushItf = outgoingPush;

    // send message to server session
    Message message = messageManagerItf.createMessage();
    MultiplexChunk multiplexChunk = newMultiplexChunk();
    multiplexChunk.setRouteId(meid.getRouteId());
    messageManagerItf.addChunk(message, BIND_CHUNK_NAME, multiplexChunk);
    try
    {
      outgoingPush.outgoingPush(message);
    }
    catch (PushException e)
    {
      try
      {
        outgoingPush.outgoingClose(session);
      }
      catch (IOException e1)
      {
        // ignored
      }
      throw new BindException(
          "Unable to send initialization message to server session : ",
          exportId, e);
    }
    synchronized (session)
    {
      while (!session.initialized)
      {
        try
        {
          session.wait();
        }
        catch (InterruptedException e)
        {
          throw new BindException(
              "Interrupted while waiting for reply message", exportId, e);
        }
      }
      if (session.bindException != null)
      {
        throw new BindException("Remote exception", exportId,
            session.bindException);
      }
    }
    return session;
  }

  /**
   * @see ChannelProtocol#export(ChannelFactory, Map)
   */
  public synchronized ExportIdentifier export(ChannelFactory channel,
      Map<String, Object> hints) throws ExportException
  {
    short newId = (short) exportedChanels.size();
    exportedChanels.put(newId, channel);
    if (lowerExportId == null)
    {
      // not yet exported, export myself
      lowerExportId = lowerLevelProtocolItf.export(this, hints);
    }
    MultiplexExportIdentifier id = new MultiplexExportIdentifier(newId,
        lowerExportId);
    return id;
  }

  /**
   * @see ChannelProtocol#unexport(ExportIdentifier)
   */
  public synchronized void unexport(ExportIdentifier exportId)
      throws InvalidExportIdentifierException
  {
    if (!(exportId instanceof MultiplexExportIdentifier))
    {
      throw new InvalidExportIdentifierException(
          "this protocol handle only MultiplexExportIdentifier", exportId);
    }
    MultiplexExportIdentifier meid = (MultiplexExportIdentifier) exportId;
    ChannelFactory channel = exportedChanels.remove(meid.getRouteId());
    if (channel == null)
    {
      throw new InvalidExportIdentifierException("Unknown export identifier",
          exportId);
    }
  }

  /**
   * @see ChannelProtocol#createExportIdentifier(Map, ExportIdentifier[])
   */
  public ExportIdentifier createExportIdentifier(Map<String, Object> info,
      ExportIdentifier[] next) throws InvalidExportIdentifierException
  {
    Object o;
    if (info == null || (o = info.get(ROUTE_ID)) == null
        || !(o instanceof Number))
    {
      throw new InvalidExportIdentifierException(
          "Unable to find or invalid routeId key in info");
    }
    if (next == null || next.length != 1)
    {
      throw new InvalidExportIdentifierException(
          "This protocol handle only one lower level protocol");
    }
    short id = ((Number) o).shortValue();
    return new MultiplexExportIdentifier(id, next[0]);
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  private MultiplexChunk newMultiplexChunk()
  {
    if (multiplexChunkFactory == null)
    {
      multiplexChunkFactory = messageManagerItf
          .getChunkFactory(MultiplexChunk.class);
    }
    return messageManagerItf.createChunk(multiplexChunkFactory);
  }

  private ExceptionChunk newExceptionChunk()
  {
    if (exceptionChunkFactory == null)
    {
      exceptionChunkFactory = messageManagerItf
          .getChunkFactory(ExceptionChunk.class);
    }
    return messageManagerItf.createChunk(exceptionChunkFactory);
  }

  // ---------------------------------------------------------------------------
  // Inner classes
  // ---------------------------------------------------------------------------

  private class ServerSessionImpl implements OutgoingPush, IncomingPush
  {

    private boolean      initialized = false;
    private boolean      closed      = false;

    private IncomingPush upperIncomingPushItf;
    private OutgoingPush lowerOutgoingPushItf;

    public ServerSessionImpl(OutgoingPush itf)
    {
      lowerOutgoingPushItf = itf;
    }

    // -------------------------------------------------------------------------
    // Implementation of the OutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see OutgoingPush#outgoingPush(Message)
     */
    public void outgoingPush(Message message) throws IOPushException
    {
      synchronized (this)
      {
        if (closed)
        {
          throw new IOPushException("Session closed");
        }
      }
      lowerOutgoingPushItf.outgoingPush(message);
    }

    /**
     * @see OutgoingPush#outgoingClose(IncomingPush)
     */
    public synchronized void outgoingClose(IncomingPush incomingPush)
        throws IOException
    {
      closed = true;
      lowerOutgoingPushItf.outgoingClose(this);
    }

    // -------------------------------------------------------------------------
    // Implementation of the IncomingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see IncomingPush#incomingPush(Message)
     */
    public void incomingPush(Message message) throws PushException
    {
      synchronized (this)
      {
        if (!initialized)
        {
          MultiplexChunk chunk = (MultiplexChunk) messageManagerItf.getChunk(
              message, BIND_CHUNK_NAME);
          if (chunk == null)
          {
            throw new IOPushException(
                "Can't find multiplex chunk in first incoming message.");
          }
          short routeId = chunk.getRouteId();
          messageManagerItf.deleteMessage(message);

          Message reply = messageManagerItf.createMessage();
          ExceptionChunk replyChunk = newExceptionChunk();
          messageManagerItf.addChunk(reply, BIND_REPLY_CHUNK_NAME, replyChunk);
          ChannelFactory channel = exportedChanels.get(routeId);
          if (channel != null)
          {
            try
            {
              upperIncomingPushItf = channel.instantiate(this);
            }
            catch (BindException e)
            {
              replyChunk.setException(e);
            }
          }
          else
          {
            replyChunk.setException(new BindException("Unknown routeId"));
          }
          try
          {
            lowerOutgoingPushItf.outgoingPush(reply);
          }
          catch (IOPushException e)
          {
            logger.log(BasicLevel.ERROR,
                "Unable to send reply to client, close lower session", e);
            try
            {
              lowerOutgoingPushItf.outgoingClose(this);
            }
            catch (IOException e1)
            {
              // ignore
            }
            return;
          }
          initialized = true;
          return;
        }
        if (closed)
        {
          throw new IOPushException("Session closed");
        }
      }
      upperIncomingPushItf.incomingPush(message);
    }

    /**
     * @see IncomingPush#incomingClosed(Object, Exception)
     */
    public synchronized void incomingClosed(Object outgoingPush,
        Exception exception)
    {
      closed = true;
      if (upperIncomingPushItf != null)
      {
        upperIncomingPushItf.incomingClosed(this, exception);
      }
    }
  }

  private class ClientSessionImpl implements OutgoingPush, IncomingPush
  {

    private boolean      initialized = false;
    private boolean      closed      = false;
    private Exception    bindException;

    private IncomingPush upperIncomingPushItf;
    private OutgoingPush lowerOutgoingPushItf;

    public ClientSessionImpl(IncomingPush itf)
    {
      upperIncomingPushItf = itf;
    }

    // -------------------------------------------------------------------------
    // Implementation of the OutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see OutgoingPush#outgoingPush(Message)
     */
    public void outgoingPush(Message message) throws IOPushException
    {
      synchronized (this)
      {
        if (closed)
        {
          throw new IOPushException("Session closed");
        }
      }
      lowerOutgoingPushItf.outgoingPush(message);
    }

    /**
     * @see OutgoingPush#outgoingClose(IncomingPush)
     */
    public synchronized void outgoingClose(IncomingPush incomingPush)
        throws IOException
    {
      closed = true;
      lowerOutgoingPushItf.outgoingClose(this);
    }

    // -------------------------------------------------------------------------
    // Implementation of the IncomingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see IncomingPush#incomingPush(Message)
     */
    public void incomingPush(Message message) throws PushException
    {
      synchronized (this)
      {
        if (!initialized)
        {
          ExceptionChunk replyChunk = (ExceptionChunk) messageManagerItf
              .getChunk(message, BIND_REPLY_CHUNK_NAME);
          if (replyChunk == null)
          {
            throw new IOPushException(
                "Can't find find reply chunk in first incoming message.");
          }
          bindException = replyChunk.getException();
          if (bindException != null)
          {
            try
            {
              lowerOutgoingPushItf.outgoingClose(this);
            }
            catch (IOException e)
            {
              // ignore
            }
          }
          messageManagerItf.deleteMessage(message);
          initialized = true;
          notify();
          return;
        }
        if (closed)
        {
          throw new IOPushException("Session closed");
        }
      }
      upperIncomingPushItf.incomingPush(message);
    }

    /**
     * @see IncomingPush#incomingClosed(Object, Exception)
     */
    public synchronized void incomingClosed(Object outgoingPush,
        Exception exception)
    {
      closed = true;
      if (upperIncomingPushItf != null)
      {
        upperIncomingPushItf.incomingClosed(this, exception);
      }
    }
  }
}
