/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing.overChannel;

import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;

import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Manage anonymous sessions
 * 
 * @author pgarcia
 */
public class AnonymousSessionManager extends HashSet<Session>
{
  /**
   * 
   */
  private static final long serialVersionUID = 2254369014602288421L;

  /**
   * field for defensive programming
   */
  private static final boolean DEFENSIVE_CHECKS  = false;

  /**
   * Inherited logger
   */
  Logger                       logger;

  /**
   * Default constructor
   * 
   * @param logger
   */
  public AnonymousSessionManager(Logger logger)
  {
    this.logger = logger;
  }

  /**
   * Add the session
   * 
   * @param session the session to add
   * @return <code>true</code> if the session was already added
   */
  @Override
  public synchronized boolean add(Session session)
  {
    if (DEFENSIVE_CHECKS == true && !(session.isAnonymous()))
      Error.bug(logger);
    if (DEFENSIVE_CHECKS == true && contains(session))
    {
      if (this.logger.isLoggable(BasicLevel.WARN))
      {
      logger.log(BasicLevel.WARN, "The anonymous session " + session
          + " is already present");
      }
    }
    if (this.logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "The anonymous session " + session
          + " is going to be added");
    }
    return super.add(session);
  }

  /**
   * Remove the session
   * 
   * @param session the session to remove
   */
  @Override
  public synchronized boolean remove(Object session)
  {
    if (DEFENSIVE_CHECKS == true && ((Session) session).isAnonymous())
      Error.bug(logger);
    if (DEFENSIVE_CHECKS == true && !contains(session))
    {
      if (this.logger.isLoggable(BasicLevel.WARN))
      {
      logger.log(BasicLevel.WARN, "The anonymous session " + session
          + " was not found");
      }
    }
    if (this.logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "The anonymous session " + session
          + " is going to be removed");
    }
    return super.remove(session);
  }

  /**
   * close all anonymous sessions
   */
  public synchronized void closeAll()
  {
    LoggerUtil.start(this.logger);
    // close anonymous session.
    Iterator<Session> iterator = iterator();
    while (iterator.hasNext())
    {
      Session session = (Session) iterator.next();
      try
      {
        this.logger.log(BasicLevel.DEBUG, "Close " + session);
        session.close();
      }
      catch (IOException e)
      {
        this.logger.log(BasicLevel.WARN,
            "Exception caught while closing lower session", e);
      }
      iterator.remove();
    }
    LoggerUtil.end(this.logger);
  }
}
