/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s):
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing.overChannel;

import java.io.IOException;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.protocol.channel.CloseChunk;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Message Passing Protocol Session
 * 
 * @author pgarcia
 */
public class Session implements IncomingPush
{
  private static final boolean                           DEFENSIVE_CHECKS           = false;

  /**
   * Chunk name of {@link ExportIdentifierChunk} send in first message of newly
   * opened channel session.
   */
  public static final String                             ID_ANNOUNCEMENT_CHUNK_NAME = "id-announcement";

  /**
   * Inherited Logger of the protocol component
   */
  protected Logger                                       logger;

  /**
   * Interface for the management of dream messages.
   */
  protected MessageManagerType                           messageManagerItf;

  /**
   * Identifier of the remote destination
   */
  ExportIdentifier                                       remoteExportIdentifier;

  /**
   * Interface on which message are sent to the lower protocol
   */
  OutgoingPush                                           outgoingPush;

  /**
   * Indicate if a close request was sent
   */
  boolean                                                closeRequestSent           = false;

  /**
   * The session which will replace this session
   */
  Session                                                replacingSession;

  /**
   * The manager which manage this session
   */
  SessionManager                                         sessionManager;

  /**
   * Chunk factory for export identifier chunk
   */
  protected ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;
  protected ChunkFactoryReference<CloseChunk>            closeChunkFactory;

  /**
   * Empty constructor
   */
  Session()
  {
  }

  Session(OutgoingPush outgoingPush, SessionManager sessionManager)
  {
    this.logger = sessionManager.logger;
    if (this.logger.isLoggable(BasicLevel.DEBUG))
      this.logger.log(BasicLevel.DEBUG, "Constructor of an anonymous session");
    this.messageManagerItf = sessionManager.messageManagerItf;
    this.outgoingPush = outgoingPush;
    this.sessionManager = sessionManager;
  }

  Session(ExportIdentifier remoteExportIdentifier, SessionManager sessionManager)
  {
    this.logger = sessionManager.logger;
    if (this.logger.isLoggable(BasicLevel.DEBUG))
      this.logger.log(BasicLevel.DEBUG, "Constructor of a session towards "
          + remoteExportIdentifier);
    this.messageManagerItf = sessionManager.messageManagerItf;
    this.remoteExportIdentifier = remoteExportIdentifier;
    this.sessionManager = sessionManager;
  }

  void init(OutgoingPush outgoingPush) throws IOPushException
  {
    LoggerUtil.start(this.logger);
    this.outgoingPush = outgoingPush;
    if (sessionManager.localExportIdentifier != null)
    {
      Message message = this.messageManagerItf.createMessage();
      ExportIdentifierChunk chunk = newExportIdentifierChunk();
      chunk.setExportIdentifier(sessionManager.localExportIdentifier);
      this.messageManagerItf.addChunk(message, ID_ANNOUNCEMENT_CHUNK_NAME,
          chunk);
      this.outgoingPush.outgoingPush(message);
    }
    LoggerUtil.end(this.logger);
  }

  void close() throws IOException
  {
    LoggerUtil.start(this.logger);
    if (DEFENSIVE_CHECKS && this.outgoingPush == null)
    {
      Error.bug(this.logger);
    }
    this.outgoingPush.outgoingClose(this);
    LoggerUtil.end(this.logger);
  }

  // -----------------------------------------------------------------------
  // Implementation of the IncomingPush interface
  // -----------------------------------------------------------------------

  /**
   * @see IncomingPush#incomingClosed(Object, Exception)
   */
  public void incomingClosed(Object outgoingPush, Exception exception)
  {

    LoggerUtil.start(logger);
    if (this.logger.isLoggable(BasicLevel.INFO))
    {
      this.logger.log(BasicLevel.INFO, "Lower " + this
          + " closed unexpectedly cause : ", exception);
    }
    if (DEFENSIVE_CHECKS && this.outgoingPush != outgoingPush)
    {
      this.logger.log(BasicLevel.ERROR, "Outgoing push caller : "
          + outgoingPush + " is different from " + this.outgoingPush);
      Error.bug(this.logger);
    }
    synchronized (sessionManager)
    {
      sessionManager.removeSession(remoteExportIdentifier, this);
    }
    LoggerUtil.end(this.logger);
  }

  /**
   * @see IncomingPush#incomingPush(Message)
   */
  public void incomingPush(Message message) throws PushException
  {
    LoggerUtil.start(this.logger);
    synchronized (sessionManager)
    { 
      if (sessionManager.closed)
      {
        throw new IOPushException("Access point has been closed.");
      }
      ExportIdentifierChunk remoteIdChunk = (ExportIdentifierChunk) this.messageManagerItf
          .getChunk(message, ID_ANNOUNCEMENT_CHUNK_NAME);
      if (remoteIdChunk != null)
      {
        treatIncomingRemotIdChunk(message, remoteIdChunk);
        this.messageManagerItf.deleteMessage(message);
        return;
      }
      CloseChunk closeChunk = (CloseChunk) this.messageManagerItf.getChunk(
          message, CloseChunk.DEFAULT_NAME);
      if (closeChunk != null)
      {
        treatIncomingCloseChunk();
        this.messageManagerItf.deleteMessage(message);
        return;
      }

      if (sessionManager.fromChunkName != null
          && remoteExportIdentifier != null)
      {
        ExportIdentifierChunk chunk = newExportIdentifierChunk();
        chunk.setExportIdentifier(remoteExportIdentifier);
        this.messageManagerItf.addChunk(message, sessionManager.fromChunkName,
            chunk);
      }
    }
    sessionManager.upperIncomingPushItf.incomingPush(message);
    LoggerUtil.end(this.logger);
  }

  private void treatIncomingRemotIdChunk(Message message,
      ExportIdentifierChunk remoteIdChunk) throws IOPushException
  {
    LoggerUtil.start(this.logger);
    setRemoteExportIdentifier(message, remoteIdChunk);
    sessionManager.sessionIdentified(remoteExportIdentifier, this);
    this.messageManagerItf.deleteMessage(message);
    LoggerUtil.end(this.logger);
  }

  private void treatIncomingCloseChunk()
  {
    LoggerUtil.start(this.logger);
    this.logger.log(BasicLevel.DEBUG, "Received a close session request for ("
        + this.remoteExportIdentifier + ")");
    sessionManager.removeAnonymousSession(this);
    try
    {
      outgoingPush.outgoingClose(this);
    }
    catch (IOException e)
    {
      this.logger.log(BasicLevel.WARN,
          "exception caught while closing lower session.", e);
    }
    LoggerUtil.end(this.logger);
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  /**
   * Get the export identifier from the received message and set the remote
   * export identifier field of this session.
   * 
   * @param message
   * @param remoteIdChunk
   * @return
   */
  private void setRemoteExportIdentifier(Message message,
      ExportIdentifierChunk remoteIdChunk)
  {
    if (DEFENSIVE_CHECKS && this.remoteExportIdentifier != null)
    {
      Error.bug(this.logger);
    }
    remoteExportIdentifier = remoteIdChunk.getExportIdentifier();
    this.messageManagerItf.removeChunk(message, ID_ANNOUNCEMENT_CHUNK_NAME);
    this.messageManagerItf.deleteChunk(remoteIdChunk);
    if (this.logger.isLoggable(BasicLevel.DEBUG))
    {
      this.logger.log(BasicLevel.DEBUG, "Received remote identifier : "
          + remoteExportIdentifier);
    }
  }

  ExportIdentifierChunk newExportIdentifierChunk()
  {
    if (exportIdChunkFactory == null)
    {
      exportIdChunkFactory = messageManagerItf
          .getChunkFactory(ExportIdentifierChunk.class);
    }
    return messageManagerItf.createChunk(exportIdChunkFactory);
  }

  /**
   * @return true if the session is anonymous
   */
  final boolean isAnonymous()
  {
    return remoteExportIdentifier == null;
  }

  /**
   * Nice format for output
   */
  @Override
  public String toString()
  {
    String res = "";
    if (remoteExportIdentifier == null)
    {
      res += "AnonymousSession@" + Integer.toHexString(hashCode()) + "(";
    }
    else
    {
      res += "Session@" + Integer.toHexString(hashCode()) + "("
          + remoteExportIdentifier + ", ";
    }
    res += "ReplacingSession : " + this.replacingSession + ")";
    return res;
  }
}