/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.pump;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.manager.ThreadPoolManager;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamLifeCycle;
import org.objectweb.dream.dreamannotation.util.DreamLifeCycleType;
import org.objectweb.fractal.julia.control.lifecycle.ChainedIllegalLifeCycleException;

/**
 * Synchronized implementation of a Pull/Push pump. This component has an
 * optional {@link org.objectweb.dream.synchro.MutexItf }client interface that
 * it uses to synchronize the execution of its run method with other components.
 * 
 * @see AbstractPumpSynchronized
 */
@DreamComponent(controllerDesc = "activeDreamPrimitive")
public class PumpSynchronized extends AbstractPumpSynchronized
{

  // ---------------------------------------------------------------------------
  // Attributes
  // ---------------------------------------------------------------------------
  
  /**
   * The <code>PushNullPolicy</code> indicates whether "null" messages that
   * are pulled must be pushed or not. Its default value is <code>false</code>.
   */
  @Attribute(argument = "pushNullPolicy")
  protected boolean pushNullPolicy            = false;
  /**
   * The <code>StopOnPushExceptionPolicy</code> indicates whether the task
   * must be stopped when a {@link org.objectweb.dream.PushException} occurs.
   * Its default value is <code>true</code>.
   */
  @Attribute(argument = "stopOnPushExceptionPolicy")
  protected boolean stopOnPushExceptionPolicy = true;
  /**
   * The <code>StopOnPullExceptionPolicy</code> indicates whether the task
   * must be stopped when a {@link org.objectweb.dream.PullException} occurs.
   * Its default value is <code>true</code>.
   */
  @Attribute(argument = "stopOnPullExceptionPolicy")
  protected boolean stopOnPullExceptionPolicy = true;
  
  /**
   * True if the pump use a pool 
   */
  @Attribute(argument = "usePool")
  protected boolean usePool                = false;
  
  /**
   * Sets the initial number of threads
   */
  @Attribute(argument = "initialNumberOfThreads")
  protected int     initialNumberOfThreads = 1;

  // -------------------------------------------------------------------------
  // Implementation of the LifeCycleController interface.
  // -------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
   */
  @LifeCycle(on = LifeCycleType.START)
  public void startFcHandler() throws IllegalLifeCycleException
  {
    if (usePool)
    {
      try
      {
        ThreadPoolManager threadPoolManager = (ThreadPoolManager) pumpTask
            .getControlItf();
        for (int i = 0; i < initialNumberOfThreads; i++)
        {
          threadPoolManager.addThread(pumpTask);
        }
      }
      catch (Exception e)
      {
        throw new ChainedIllegalLifeCycleException(e, null,
            "An error occurred while retrieving task controller interface");
      }
    }
  }

  /**
   * @see org.objectweb.dream.AbstractComponent#beforeFirstStart(Component)
   */
  @DreamLifeCycle(on=DreamLifeCycleType.FIRST_START)
  protected void beforeFirstStart(Component componentItf)
      throws IllegalLifeCycleException
  {
    try
    {
      if (usePool)
      {
        Map<String, Object> m = new HashMap<String, Object>();
        m.put("thread", "pool");
        Util.addTask(componentItf, pumpTask, m);
        // Note that threads cannot be added since the tasks is not started (so
        // it does not have any controller)
        // Threads are added in the startFc method.
      }
      else
      {
        Util.addTask(componentItf, pumpTask, null);
      }
    }
    catch (Exception e)
    {
      throw new ChainedIllegalLifeCycleException(e, componentItf,
          "Unable to register the pump task");
    }
  }
}
