/**
 * Dream
 * Copyright (C) 2003 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.queue;

import java.util.LinkedList;
import java.util.List;
import java.util.Collections;


import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.queue.buffer.BufferMonitoring;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;

/**
 * High performance implementation of a Push/Pull queue. This queue has the
 * following properties :
 * <ul>
 * <li>FIFO</li>
 * <li>Thread safe</li>
 * <li>Has no maximum capacity</li>
 * <li>{@link #pull} is blocking if queue is empty.</li>
 * </ul>
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
@Provides(interfaces = {
    @Interface(name = "out-pull", signature = Pull.class),
    @Interface(name = "in-push", signature = Push.class)})
public class PushPullQueueNonBlockingFast 
    implements
      Push,
      Pull,
      BufferMonitoring
{

  // ------------------------------------------------------------------------
  // Fields.
  // ------------------------------------------------------------------------

  protected List<Message> list        = Collections
                                          .synchronizedList(new LinkedList<Message>());
  
  // ------------------------------------------------------------------------
  // Implementation of the Push interface.
  // ------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    try
    {
      list.add(message);
    }
    catch (UnsupportedOperationException e)
    {
      throw new PushException(
          "The class of the specified element prevents it from being added to this list");
    }
    catch (NullPointerException e)
    {
      throw new PushException(
          "The specified element is null and this list does not support null elements");
    }
    catch (IllegalArgumentException e)
    {
      throw new PushException(
          "Some aspect of this element prevents it from being added to this list");
    }
  }

  // ------------------------------------------------------------------------
  // Implementation of the Pull interface.
  // ------------------------------------------------------------------------

  /**
   * @see Pull#pull()
   */
  public Message pull() throws PullException
  {
    if (list.size() > 0)
    {
      try
      {
        Message message = list.remove(0);
        return message;
      }
      catch (IndexOutOfBoundsException e)
      {
        throw new PullException(
            "The index is out of range (index < 0 || index >= size())");
      }
    }
    return null;
  }

  // ------------------------------------------------------------------------
  // Implementation of the BufferMonitoring interface
  // ------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.buffer.BufferMonitoring#getCurrentSize()
   */
  public int getCurrentSize()
  {
    return list.size();
  }
}