/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Romain Lenglet
 * Contributor(s):
 */

package org.objectweb.dream.queue.buffer;

import java.util.Iterator;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import org.objectweb.dream.message.Message;
import org.objectweb.dream.queue.MessageComparableKeyManager;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;

/**
 * A buffer in which stored messages are available to {@link #get() get} or
 * {@link #remove() remove} in the order of the "key" that is calculated for
 * each message by the {@link MessageComparableKeyManager} component bound to
 * this component when {@link #add(Message) adding} each message. Messages that
 * have the "highest" key are retrieved first.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class BufferSorting extends AbstractBuffer
{
  // ------------------------------------------------------------------------
  // Client interface.
  // ------------------------------------------------------------------------

  /**
   * Used to calculate a key for each added message.
   */
  @Requires(name = MessageComparableKeyManager.ITF_NAME)
  protected MessageComparableKeyManager    messageComparableKeyManager;

  // ------------------------------------------------------------------------
  // Fields
  // ------------------------------------------------------------------------

  /**
   * The messages stored in this buffer, indexed by their key.
   */
  protected SortedMap<Comparable, Message> messages               = new TreeMap<Comparable, Message>();

  // ---------------------------------------------------------------------------
  // Implementation of the AbstractBuffer methods.
  // ---------------------------------------------------------------------------

  protected void doAdd(Message message)
  {
    /* Calculate the key for the newly added message. */
    Comparable key = messageComparableKeyManager.createComparableKey();
    messageComparableKeyManager.fillComparableKey(key, message);

    messages.put(key, message);

    incrementAvailableMessagesCount(1);
    incrementStoredMessagesCount(1);
  }

  protected Message doGet()
  {
    return messages.get(messages.lastKey());
  }

  protected Message doRemove()
  {
    /* Remove message. */
    Comparable key = messages.lastKey();
    Message message = messages.get(key);

    /* Destroy key. */
    messageComparableKeyManager.deleteComparableKey(key);

    incrementAvailableMessagesCount(-1);
    incrementStoredMessagesCount(-1);

    return message;
  }

  protected Message doRemoveAll()
  {
    int size = messages.size();

    if (size == 0)
    {
      return null;
    }
    Message aggregatedMessage = messageManagerItf.createMessage();

    Iterator messagesIterator = messages.entrySet().iterator();
    while (messagesIterator.hasNext())
    {
      Map.Entry entry = (Map.Entry) messagesIterator.next();

      Comparable key = (Comparable) entry.getKey();
      messageComparableKeyManager.deleteComparableKey(key);

      Message message = (Message) entry.getValue();
      messageManagerItf.addSubMessage(aggregatedMessage, message);
    }

    messages.clear();
    incrementAvailableMessagesCount(-size);
    incrementStoredMessagesCount(-size);
    return aggregatedMessage;
  }
  
}
