/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s):
 */

package org.objectweb.dream.queue.handler.incoming;


import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.queue.buffer.BufferAdd;
import org.objectweb.dream.queue.buffer.BufferRemove;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;

/**
 * Abstract implementation of an incoming handler. It implements the
 * {@link Push push} interface.
 * <p>
 * An incoming handler aims at treating incoming messages coming from a
 * {@link org.objectweb.dream.queue.AbstractBuffer buffer}
 * <p>
 * Incoming handler developers can inherit this class. They must implement one
 * method:
 * <ul>
 * <li><code>push(message)</code> push a message to the related buffer. </li>
 * </p>
 */
@DreamComponent()
@Provides(interfaces = @Interface(name = "in-push", signature = Push.class))
public abstract class AbstractIncomingHandler 
    implements
      Push
{

  // ------------------------------------------------------------------------
  // Client interfaces.
  // ------------------------------------------------------------------------

  /**
   * The buffer from which pulled messages are removed.
   */
  @Requires(name = BufferAdd.ITF_NAME)
  protected BufferAdd          bufferAdd;

  /**
   * Used to delete dropped messages in {@link IncomingHandlerDropping}.
   */
  @Requires(name = "message-manager")
  protected MessageManagerType messageManagerItf;

  /**
   * The buffer from which message may be removed. Used in
   * {@link IncomingHandlerRemoving}.
   */
  @Requires(name = BufferRemove.ITF_NAME, contingency = Contingency.OPTIONAL)
  protected BufferRemove       bufferRemove;

  // ------------------------------------------------------------------------
  // Method that should be overridden by concrete sub classes.
  // ------------------------------------------------------------------------

  /**
   * @see Push#push(Message message)
   */
  public abstract void push(Message message) throws PushException;

}
