/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.queue.keyed.buffer;

import java.util.Iterator;
import java.util.TreeMap;

import org.objectweb.dream.message.Message;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.Logger;

/**
 * Buffer that uses keys to sort messages. The key corresponding to a message is
 * explicitely passed as a parameter, when adding the message using the
 * {@link #add(Message, Object)} method. The {@link #doRemove(Object)} method
 * removes the message with the specified key. If no key is specified in the
 * <code>doRemove()</code> method, this latter returns the message with the
 * highest key.
 * <p>
 * <b>Note: </b>Keys must implement the {@link java.lang.Comparable}interface.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class BufferKeyedAddKeyedRemove
    extends
      AbstractBufferKeyedAddKeyedRemove
{

  // ---------------------------------------------------------------------------
  // Fields.
  // ---------------------------------------------------------------------------

  TreeMap<Object, Message> queue = new TreeMap<Object, Message>();
  
  // --------------------------------------------------------------------------
  // Services interfaces
  // --------------------------------------------------------------------------  
  
  /**
   * Component reference
   */
  @Service
  Component weaveableC;
  
  /**
   * Logger of the component
   */ 
  @DreamMonolog()
  protected Logger                   logger;  

  // ------------------------------------------------------------------------
  // Overridden methods.
  // ------------------------------------------------------------------------

  protected void doAdd(Message message, Object key)
  {
    if (key == null)
    {
      Error.error("Key is null", logger);
    }
    queue.put(key, message);
    incrementAvailableMessagesCount(1, key);
    incrementStoredMessagesCount(1, key);
  }

  protected Message doGet(Object key)
  {
    Message message;
    if (key == null)
    {
      message = queue.get(queue.lastKey());
    }
    else
    {
      message = queue.get(key);
    }
    return message;
  }

  protected Message doRemove(Object key)
  {
    Message message;
    if (key == null)
    {
      message = queue.remove(queue.lastKey());
    }
    else
    {
      message = queue.remove(key);
    }
    if (message != null)
    {
      incrementAvailableMessagesCount(-1, key);
      incrementStoredMessagesCount(-1, key);
    }
    return message;
  }

  protected Message doRemoveAll()
  {
    int size = queue.size();
    Iterator<Message> iter = queue.values().iterator();
    Message message = messageManagerItf.createMessage();
    while (iter.hasNext())
    {
      messageManagerItf.addSubMessage(message, iter.next());
      iter.remove();
    }
    incrementAvailableMessagesCount(-size, null);
    incrementStoredMessagesCount(-size, null);
    return message;
  }
}