/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre GARCIA
 * Contributor(s):
 */

package org.objectweb.dream.queue.keyed.buffer.buffer;

import java.util.Iterator;
import java.util.TreeMap;

import org.objectweb.dream.message.Message;
import org.objectweb.dream.queue.buffer.BufferFIFO;
import org.objectweb.dream.queue.keyed.buffer.AbstractBufferKeyedAddKeyedRemove;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.Logger;

/**
 * Buffer that uses keys to sort messages. Each elements of the buffer is a FIFO
 * List of messages. The key corresponding to a message is explicitely passed as
 * a parameter, when adding the message using the {@link #add(Message, Object)}
 * method. The {@link #doRemove(Object)} method removes the message with the
 * specified key. If no key is specified in the <code>doRemove()</code>
 * method, this latter returns the message with the highest key.
 * <p>
 * <b>Note: </b>Keys must implement the {@link java.lang.Comparable}interface.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class FIFOBuffersKeyedAddKeyedRemove
    extends
      AbstractBufferKeyedAddKeyedRemove
{

  // ---------------------------------------------------------------------------
  // Fields.
  // ---------------------------------------------------------------------------

  TreeMap<Object, BufferFIFO> queue = new TreeMap<Object, BufferFIFO>();

  // --------------------------------------------------------------------------
  // Services interfaces
  // --------------------------------------------------------------------------

  /**
   * Component reference
   */
  @Service
  Component                   weaveableC;

  /**
   * Logger of the component
   */
  @DreamMonolog()
  protected Logger            logger;

  // --------------------------------------------------------------------------
  // Overridden methods.
  // --------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.keyed.buffer.AbstractBufferKeyedAddKeyedRemove#hasAvailableMessage(java.lang.Object)
   */
  protected boolean hasAvailableMessage(Object key)
  {
    if (key == null)
    {
      Error.error("Key is null", logger);
    }
    if (queue.get(key) == null)
    {
      return false;
    }
    else
    {
      return queue.get(key).hasAvailableMessage();
    }
  }

  /**
   * @throws InterruptedException
   * @see org.objectweb.dream.queue.keyed.buffer.AbstractBufferKeyedAddKeyedRemove#doAdd(org.objectweb.dream.message.Message,
   *      java.lang.Object)
   */
  protected void doAdd(Message message, Object key)
  {
    if (key == null)
    {
      Error.error("Key is null", logger);
    }
    BufferFIFO buffer;
    if (!queue.containsKey(key))
    {
      buffer = queue.get(key);
    }
    else
    {
      buffer = new BufferFIFO();
      queue.put(key, buffer);
    }
    buffer.tryAdd(message);
    incrementAvailableMessagesCount(1, key);
    incrementStoredMessagesCount(1, key);
  }

  /**
   * @see org.objectweb.dream.queue.keyed.buffer.AbstractBufferKeyedAddKeyedRemove#doGet(java.lang.Object)
   */
  protected Message doGet(Object key)
  {
    Message message;
    if (key == null)
    {
      Error.error("Key is null", logger);
      message = queue.get(queue.lastKey()).tryGet();
    }
    else
    {
      message = queue.get(key).tryGet();
    }
    return message;
  }

  /**
   * @see org.objectweb.dream.queue.keyed.buffer.AbstractBufferKeyedAddKeyedRemove#doRemove(java.lang.Object)
   */
  protected Message doRemove(Object key)
  {
    Message message = null;
    if (key == null)
    {
      Error.error("Key is null", logger);
    }
    else
    {
      BufferFIFO  buffer = queue.get(key);
      if (buffer != null){
        message = buffer.tryRemove();
      }
    }
    if (message != null)
    {
      incrementAvailableMessagesCount(-1, key);
      incrementStoredMessagesCount(-1, key);
    }
    return message;
  }

  /**
   * @see org.objectweb.dream.queue.keyed.buffer.AbstractBufferKeyedAddKeyedRemove#doRemoveAll()
   */
  protected Message doRemoveAll()
  {
    Iterator<BufferFIFO> iter = queue.values().iterator();
    Message message = messageManagerItf.createMessage();
    while (iter.hasNext())
    {
      BufferFIFO buffer = iter.next();
      int size = buffer.getCurrentSize();
      messageManagerItf.addSubMessage(message, buffer.removeAll());
      iter.remove();
      incrementAvailableMessagesCount(-size, null);
      incrementStoredMessagesCount(-size, null);
    }
    return message;
  }
}