/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;

import java.lang.ref.Reference;

import org.objectweb.dream.pool.Recyclable;
import org.objectweb.dream.dreamannotation.DreamComponent;

/**
 * Abstract implementation of a semaphore relying on queued wait nodes.
 * <p>
 * <b>Note: Inspired by Doug Lea's implementation. </b>
 */
@DreamComponent()
public class AbstractSemaphoreQueuedImpl extends Semaphore
{
  protected final WaitQueue wq;

  AbstractSemaphoreQueuedImpl(WaitQueue q)
  {
    wq = q;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the SemaphoreItf interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#acquire()
   */
  public void acquire() throws InterruptedException
  {
    if (Thread.interrupted())
    {
      throw new InterruptedException();
    }
    if (precheck())
    {
      return;
    }
    WaitNode w = new WaitNode();
    w.doWait(this);
  }

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#attempt(long)
   */
  public boolean attempt(long msecs) throws InterruptedException
  {
    if (Thread.interrupted())
    {
      throw new InterruptedException();
    }
    if (precheck())
    {
      return true;
    }
    if (msecs <= 0)
    {
      return false;
    }

    WaitNode w = new WaitNode();
    return w.doTimedWait(this, msecs);
  }

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#release()
   */
  public void release()
  {
    for (;;)
    {
      WaitNode w = getSignallee();
      if (w == null)
      {
        return; // no one to signal
      }
      if (w.signal())
      {
        return; // notify if still waiting, else skip
      }
    }
  }

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#release(long)
   */
  public void release(long n)
  {
    if (n < 0)
    {
      throw new IllegalArgumentException("Negative argument");
    }

    for (long i = 0; i < n; ++i)
    {
      release();
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods.
  // ---------------------------------------------------------------------------

  protected synchronized boolean precheck()
  {
    boolean pass = (permits > 0);
    if (pass)
    {
      --permits;
    }
    return pass;
  }

  protected synchronized boolean recheck(WaitNode w)
  {
    boolean pass = (permits > 0);
    if (pass)
    {
      --permits;
    }
    else
    {
      wq.insert(w);
    }
    return pass;
  }

  protected synchronized WaitNode getSignallee()
  {
    WaitNode w = wq.extract();
    if (w == null)
    {
      ++permits; // if none, inc permits for new arrivals
    }
    return w;
  }

  // ---------------------------------------------------------------------------
  // Inner class.
  // ---------------------------------------------------------------------------

  /**
   * Base class for internal queue classes for semaphores, etc. Relies on
   * subclasses to actually implement queue mechanics
   */

  public abstract static class WaitQueue
  {

    protected abstract void insert(WaitNode w);// assumed not to block

    protected abstract WaitNode extract(); // should return null if empty
  }
  
  /**
   * An element of the queue.
   */
  protected final class WaitNode implements Recyclable
  {
    boolean                          waiting = true;
    WaitNode                         next    = null;
    Reference< ? extends Recyclable> recyclableReference;

    protected synchronized boolean signal()
    {
      boolean signalled = waiting;
      if (signalled)
      {
        waiting = false;
        notify();
      }
      return signalled;
    }

    protected synchronized boolean doTimedWait(
        AbstractSemaphoreQueuedImpl sem, long msecs)
        throws InterruptedException
    {
      if (sem.recheck(this) || !waiting)
      {
        return true;
      }
      else if (msecs <= 0)
      {
        waiting = false;
        return false;
      }
      else
      {
        long waitTime = msecs;
        long start = System.currentTimeMillis();

        try
        {
          for (;;)
          {
            wait(waitTime);
            if (!waiting)
            {// definitely signaled
              return true;
            }
            waitTime = msecs - (System.currentTimeMillis() - start);
            if (waitTime <= 0)
            { // timed out
              waiting = false;
              return false;
            }
          }
        }
        catch (InterruptedException ex)
        {
          if (waiting)
          { // no notification
            waiting = false; // invalidate for the signaler
            throw ex;
          }
          // thread was interrupted after it was notified
          Thread.currentThread().interrupt();
          return true;
        }
      }
    }

    protected synchronized void doWait(AbstractSemaphoreQueuedImpl sem)
        throws InterruptedException
    {
      if (!sem.recheck(this))
      {
        try
        {
          while (waiting)
          {
            wait();
          }
        }
        catch (InterruptedException ex)
        {
          if (waiting)
          { // no notification
            waiting = false; // invalidate for the signaler
            throw ex;
          }
          // thread was interrupted after it was notified
          Thread.currentThread().interrupt();
          return;
        }
      }
    }

    // ---------------------------------------------------------------------------
    // Implementation of the Recyclable interface.
    // ---------------------------------------------------------------------------

    /**
     * @see Recyclable#recycle()
     */
    public void recycle()
    {
      waiting = true;
      next = null;
    }

    /**
     * @see Recyclable#getReference()
     */
    public Reference< ? extends Recyclable> getReference()
    {
      return recyclableReference;
    }

    /**
     * @see Recyclable#setReference(Reference)
     */
    public void setReference(
        Reference< ? extends Recyclable> recyclableReference)
    {
      this.recyclableReference = recyclableReference;
    }

  }
}
