/**
 * DREAM
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;

import org.objectweb.fractal.fraclet.annotation.annotations.Interface;

/**
 * A condition component aima at synchronizing the execution of several
 * components. ConditionItf components use {@link MutexItf}components.
 * <p>
 * <b>Note: Inspired by Doug Lea's implementation. </b>
 * 
 * @see MutexItf
 */
@Interface(name = ConditionItf.ITF_NAME)
public interface ConditionItf
{
  /**
   * The commonly used name to refer to this interface.
   */
  String ITF_NAME = "condition";

  /**
   * Wait for notification. This operation at least momentarily releases the
   * mutex. The mutex is always held upon return, even if interrupted.
   * 
   * @exception InterruptedException if the thread was interrupted before or
   *              during the wait. However, if the thread is interrupted after
   *              the wait but during mutex re-acquisition, the interruption is
   *              ignored, while still ensuring that the currentThread's
   *              interruption state stays true, so can be probed by callers.
   */
  void await() throws InterruptedException;

  /**
   * Wait for at most msecs for notification. This operation at least
   * momentarily releases the mutex. The mutex is always held upon return, even
   * if interrupted.
   * 
   * @param msecs The time to wait. A value less than or equal to zero causes a
   *          momentarily release and re-acquire of the mutex, and always
   *          returns false.
   * @return false if at least msecs have elapsed upon resumption; else true. A
   *         false return does NOT necessarily imply that the thread was not
   *         notified. For example, it might have been notified after the time
   *         elapsed but just before resuming.
   * @exception InterruptedException if the thread was interrupted before or
   *              during the wait.
   */
  boolean timedwait(long msecs) throws InterruptedException;

  /**
   * Notify a waiting thread. If one exists, a non-interrupted thread will
   * return normally (i.e., not via InterruptedException) from await or
   * timedwait.
   */
  void signal();

  /** Notify all waiting threads * */
  void signalAll();
}