/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;

import org.objectweb.fractal.fraclet.annotation.annotations.Interface;

/**
 * A read-write lock component aims at synchronizing the execution of several
 * reader and writer components. Multiple reader components can concurrently
 * read, so long as there are no writers. Only one writer component can write at
 * a given time.
 * <p>
 * <b>Note: This interface is freely inspired by Doug Lea's
 * <code>ReadWriteLockItf</code> interface. </b>
 */
@Interface(name=ReadWriteLockItf.ITF_NAME)
public interface ReadWriteLockItf
{

  /** The default name of the {@link ReadWriteLockItf }interface. */
  String ITF_NAME = "read-write-lock";

  /**
   * Gets the read lock.
   * 
   * @throws InterruptedException if the thread is interrupted while trying to
   *           acquire the lock.
   */
  void acquireRead() throws InterruptedException;

  /**
   * Releases the read lock.
   */
  void releaseRead();

  /**
   * Attempts to get the read lock.
   * 
   * @param msecs the number of milleseconds to wait. An argument less than or
   *          equal to zero means not to wait at all. However, this may still
   *          require access to a synchronization lock, which can impose
   *          unbounded delay if there is a lot of contention among threads.
   * @return true if acquired.
   * @throws InterruptedException if the thread is interrupted while waiting for
   *           the lock.
   */
  boolean attemptRead(long msecs) throws InterruptedException;

  /**
   * Gets the write lock.
   * 
   * @throws InterruptedException if the thread is interrupted while trying to
   *           acquire the lock.
   */
  void acquireWrite() throws InterruptedException;

  /**
   * Releases the write lock.
   */
  void releaseWrite();

  /**
   * Attempts to get the write lock.
   * 
   * @param msecs the number of milleseconds to wait. An argument less than or
   *          equal to zero means not to wait at all. However, this may still
   *          require access to a synchronization lock, which can impose
   *          unbounded delay if there is a lot of contention among threads.
   * @return true if acquired.
   * @throws InterruptedException if the thread is interrupted while waiting for
   *           the lock.
   */
  boolean attemptWrite(long msecs) throws InterruptedException;

}