/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;


import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.dream.dreamannotation.DreamComponent;

/**
 * Basic implementation of the {@link SemaphoreItf}interface.
 * <p>
 * <b>Note: Inspired by Doug Lea's implementation. </b>
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class Semaphore 
    implements
      SemaphoreItf
{

  /** current number of available permits * */
  @Attribute(argument = "initialPermits")
  protected long permits;

  // ---------------------------------------------------------------------------
  // Implementation of the SemaphoreItf interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#acquire()
   */
  public void acquire() throws InterruptedException
  {
    if (Thread.interrupted()) throw new InterruptedException();
    synchronized (this)
    {
      try
      {
        while (permits <= 0)
          wait();
        --permits;
      }
      catch (InterruptedException ex)
      {
        notify();
        throw ex;
      }
    }
  }

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#attempt(long)
   */
  public boolean attempt(long msecs) throws InterruptedException
  {
    if (Thread.interrupted()) throw new InterruptedException();

    synchronized (this)
    {
      if (permits > 0)
      {
        --permits;
        return true;
      }
      else if (msecs <= 0)
        return false;
      else
      {
        try
        {
          long startTime = System.currentTimeMillis();
          long waitTime = msecs;

          for (;;)
          {
            wait(waitTime);
            if (permits > 0)
            {
              --permits;
              return true;
            }
            waitTime = msecs - (System.currentTimeMillis() - startTime);
            if (waitTime <= 0)
            {
              return false;
            }
          }
        }
        catch (InterruptedException ex)
        {
          notify();
          throw ex;
        }
      }
    }
  }

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#release()
   */
  public synchronized void release()
  {
    ++permits;
    notify();
  }

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#release(long)
   */
  public synchronized void release(long n)
  {
    if (n < 0) throw new IllegalArgumentException("Negative argument");

    permits += n;
    for (long i = 0; i < n; ++i)
      notify();
  }

  /**
   * @see org.objectweb.dream.synchro.SemaphoreItf#permits()
   */
  public synchronized long permits()
  {
    return permits;
  }

//  // ---------------------------------------------------------------------------
//  // Implementation of the SemaphoreAttributeController interface.
//  // ---------------------------------------------------------------------------
//
//  /**
//   * @see SemaphoreAttributeController#setInitialPermits(long)
//   */
//  public void setInitialPermits(long initialPermits)
//  {
//    permits = initialPermits;
//  }
//
//  /**
//   * @see org.objectweb.dream.synchro.SemaphoreAttributeController#getInitialPermits()
//   */
//  public long getInitialPermits()
//  {
//    return permits;
//  }
}
